"""Search for GRB-GW coincidences with ligo-raven."""
import ligo.raven.gracedb_events
import ligo.raven.search

from ..celery import app
from . import gracedb


def coincidence_search(gracedb_id, alert_object):
    """Perform ligo-raven search for coincidences.
    The ligo.raven.search.search method applies EM_COINC label on its own.

    Parameters
    ----------
    gracedb_id: str
        ID of the trigger used by GraceDb
    alert_object: dict
        lvalert['object']
    """
    return (
        search.s(gracedb_id, alert_object)
        |
        add_exttrig_to_superevent.s(gracedb_id)
    )


@app.task(shared=False)
def search(gracedb_id, alert_object):
    """Perform ligo-raven search for coincidences.
    The ligo.raven.search.search method applies EM_COINC label on its own.

    Parameters
    ----------
    gracedb_id: str
        ID of the trigger used by GraceDb
    alert_object: dict
        lvalert['object']
    Returns
    -------
        list with the dictionaries of related gracedb events
    """
    if alert_object.get('superevent_id'):
        cls = ligo.raven.gracedb_events.SE
    else:
        cls = ligo.raven.gracedb_events.ExtTrig
    event = cls(gracedb_id, gracedb=gracedb.client)
    return ligo.raven.search.search(event, -5, 5, gracedb=gracedb.client)


@app.task(shared=False)
def add_exttrig_to_superevent(raven_search_results, gracedb_id):
    """Add external trigger to the list of em_events after
    ligo.raven.search.search finds a coincidence

    Parameters
    ----------
    raven_search_results: list
        list of dictionaries of each related gracedb trigger
    gracedb_id: str
        ID of either a superevent or external trigger
    """
    # First determine whether the gracedb_id is for a superevent or exttrig
    if gracedb_id.startswith('E'):
        for superevent in raven_search_results:
            superevent_id = superevent['superevent_id']
            gracedb.client.addEventToSuperevent(superevent_id, gracedb_id)
    else:
        for exttrig in raven_search_results:
            exttrig_id = exttrig['graceid']
            gracedb.client.addEventToSuperevent(gracedb_id, exttrig_id)



from unittest.mock import patch

from ..tasks import orchestrator
from . import resource_json


# @patch('gwcelery.tasks.circulars.create_circular')
# @patch('gwcelery.tasks.skymaps.annotate_fits')
# @patch('gwcelery.tasks.bayestar.bayestar')
# @patch('gwcelery.tasks.gcn.send.delay')
# def test_handle_voevent(mock_send, mock_bayestar, mock_annotate_fits,
#                         mock_create_circular):
#     """Test dispatch of a VOEvent message."""
#     # Test LVAlert payload.
#     alert = resource_json(__name__, 'data/lvalert_voevent.json')
#
#     # text = alert['object']['text']
#
#     # Run function under test
#     orchestrator.handle(alert)
#
#     # Check that the correct tasks were dispatched.
#     mock_annotate_fits.assert_not_called()
#     mock_bayestar.assert_not_called()
#     mock_create_circular.assert_not_called()
#     # FIXME: temporarily disable sending GCNs as per P. Brady request
#     mock_send.assert_not_called()  # mock_send.assert_called_once_with(text)


@patch('gwcelery.tasks.raven.coincidence_search')
def test_handle_superevent_creation(mock_raven_coincidence_search):
    """Test dispatch of an LVAlert message for a superevent creation."""
    # Test LVAlert payload.
    alert = resource_json(__name__, 'data/lvalert_superevent_creation.json')

    # Run function under test
    orchestrator.handle_superevents_externaltriggers(alert)

    # Check that the correct tasks were dispatched.
    mock_raven_coincidence_search.assert_called_once_with('S180616h',
                                                          alert['object'])



import json
import logging
import os
import socket
from unittest.mock import patch
from xml.etree.ElementTree import XML

import pkg_resources
from pyxmpp2.exceptions import DNSError
import pytest

from ..tasks import lvalert


@pytest.fixture
def netrc_lvalert2(tmpdir):
    path = str(tmpdir / 'netrc')
    with open(path, 'w') as f:
        print('machine', 'lvalert2.invalid', file=f)
        print('login', 'albert.einstein', file=f)
        print('password', 'foobar', file=f)
    with patch.dict(os.environ, NETRC=path):
        yield


def test_lvalert_netrc_does_not_match(netrc_lvalert2):
    """Test that we get the correct error message when the lvalert host is
    not in the netrc file."""
    with pytest.raises(RuntimeError, match='No matching netrc entry found'):
        lvalert.listen()


@pytest.fixture
def netrc_lvalert(tmpdir):
    path = str(tmpdir / 'netrc')
    with open(path, 'w') as f:
        print('machine', 'lvalert.invalid', file=f)
        print('login', 'albert.einstein', file=f)
        print('password', 'foobar', file=f)
    with patch.dict(os.environ, NETRC=path):
        yield


try:
    socket.gethostbyname('lvalert.invalid')
except socket.error:
    resolves_invalid_urls = False
else:
    resolves_invalid_urls = True


@pytest.mark.skipif(
    resolves_invalid_urls, reason='your DNS server is resolving invalid '
    'hostnames (maybe a home internet provider that provides a default '
    'landing page?)')
@pytest.mark.enable_socket
def test_lvalert_constructor(netrc_lvalert):
    """Test that we at least attempt to connect to a non-existent URL."""
    with pytest.raises(DNSError):
        lvalert.listen()


@pytest.fixture
def fake_lvalert():
    xml = XML(pkg_resources.resource_string(__name__, 'data/lvalert_xmpp.xml'))
    entry = xml.find('.//ns2:entry', lvalert.ns)
    return xml, entry


@patch('gwcelery.tasks.superevents.handle.run')
def test_handle_messages(mock_superevents_handle,
                         netrc_lvalert, fake_lvalert):
    """Test handling an LVAlert message that originates from the configured
    GraceDb server."""
    xml, entry = fake_lvalert

    # Manipulate alert content
    alert = json.loads(entry.text)
    alert['object']['self'] = \
        alert['object']['self'].replace('gracedb.ligo.org', 'gracedb.invalid')
    entry.text = json.dumps(alert)

    # Run function under test
    lvalert._handle_messages(xml)
    mock_superevents_handle.assert_called_once()


@patch('gwcelery.tasks.superevents.handle.run')
def test_handle_messages_wrong_server(mock_superevents_handle,
                                      netrc_lvalert, fake_lvalert, caplog):
    """Test handling an LVAlert message that originates from a GraceDb server
    other than the configured GraceDb server. It should be ignored."""
    xml, entry = fake_lvalert

    # Manipulate alert content
    alert = json.loads(entry.text)
    alert['object']['self'] = \
        alert['object']['self'].replace('gracedb.ligo.org', 'gracedb2.invalid')
    entry.text = json.dumps(alert)

    # Run function under test
    lvalert._handle_messages(xml)

    # Run function under test
    caplog.set_level(logging.WARNING)
    lvalert._handle_messages(xml)
    record, *_ = caplog.records
    assert record.message == ('ignoring LVAlert message because it is '
                              'intended for GraceDb server '
                              'https://gracedb2.invalid/api/, but we are set '
                              'up for server https://gracedb.invalid/api/')
    mock_superevents_handle.assert_not_called()


@patch('gwcelery.tasks.superevents.handle.run')
def test_handle_messages_no_self_link(mock_superevents_handle,
                                      netrc_lvalert, fake_lvalert, caplog):
    """Test handling an LVAlert message that does not identify the GraceDb
    server of origin. It should be rejected."""
    xml, entry = fake_lvalert

    # Manipulate alert content
    alert = json.loads(entry.text)
    del alert['object']['self']
    entry.text = json.dumps(alert)

    # Run function under test
    caplog.set_level(logging.ERROR)
    lvalert._handle_messages(xml)
    record, = caplog.records
    assert 'LVAlert message does not contain an API URL' in record.message
    mock_superevents_handle.assert_not_called()



from unittest import mock

import pytest

from .. import app
from .. import condor


@pytest.mark.parametrize('subcommand,extra_args', [['rm', ()],
                                                   ['hold', ()],
                                                   ['release', ()],
                                                   ['q', ('-nobatch',)]])
@mock.patch('os.execvp', side_effect=SystemExit(0))
def test_condor_subcommand(mock_execvp, subcommand, extra_args):
    """Test all trivial Condor subcommands."""
    try:
        app.start(['gwcelery', 'condor', subcommand])
    except SystemExit as e:
        assert e.code == 0

    cmd = 'condor_' + subcommand
    mock_execvp.assert_called_once_with(
        cmd, (cmd,) + extra_args + condor.CONSTRAINTS)


@mock.patch('subprocess.check_output', return_value=b'<classads></classads>')
@mock.patch('os.execvp', side_effect=SystemExit(0))
def test_condor_submit_not_yet_running(mock_execvp, mock_check_output):
    """Test starting the Condor job."""
    try:
        app.start(['gwcelery', 'condor', 'submit'])
    except SystemExit as e:
        assert e.code == 0

    mock_check_output.assert_called_once_with(
        ('condor_q', '-xml') + condor.CONSTRAINTS)
    mock_execvp.assert_called_once_with(
        'condor_submit', ('condor_submit', condor.SUBMIT_FILE))


@mock.patch('subprocess.check_output',
            return_value=b'<classads><c></c></classads>')
@mock.patch('os.execvp', side_effect=SystemExit(0))
def test_condor_submit_already_running(mock_execvp, mock_check_output):
    """Test that we don't start the condor jobs if they are already running."""
    try:
        app.start(['gwcelery', 'condor', 'submit'])
    except SystemExit as e:
        assert e.code == 1

    mock_check_output.assert_called_once_with(
        ('condor_q', '-xml') + condor.CONSTRAINTS)
    mock_execvp.assert_not_called()



from unittest.mock import patch

from pkg_resources import resource_string

from ..tasks.gcn import external_triggers


@patch('gwcelery.tasks.gracedb.create_event')
def test_handle_create_event(mock_create_event):
    text = resource_string(__name__, 'data/fermi_grb_gcn.xml')
    external_triggers.handle(payload=text)
    mock_create_event.assert_called_once_with(filecontents=text,
                                              search='GRB',
                                              pipeline='Fermi',
                                              group='Test')


@patch('gwcelery.tasks.gracedb.get_events', return_value=[{'graceid': 'E1'}])
@patch('gwcelery.tasks.gracedb.replace_event')
def test_handle_replace_event(mock_replace_event, mock_get_events):
    text = resource_string(__name__, 'data/fermi_grb_gcn.xml')
    external_triggers.handle(payload=text)
    mock_replace_event.assert_called_once_with('E1', text)



import gzip
import io
from unittest.mock import patch

from astropy.table import Table
from ligo.gracedb import rest
import numpy as np
import pkg_resources
import pytest

from ..tasks import skymaps


def resource_unicode(*args, **kwargs):
    with open(pkg_resources.resource_filename(*args, **kwargs), 'r') as f:
        return f.read()


@pytest.fixture
def toy_fits_filecontents():
    """Generate the binary contents of a toy FITS file."""
    bytesio = io.BytesIO()
    table = Table([[1, 2, 3], [4, 5, 6]], names=['foo', 'bar'])
    table.meta['comment'] = 'This is a comment.'
    table.meta['history'] = 'This is a history line. <This should be escaped.>'
    table.meta['ORDERING'] = 'NESTED'
    with gzip.GzipFile(fileobj=bytesio, mode='wb') as f:
        table.write(f, format='fits')
    return bytesio.getvalue()


@pytest.fixture
def toy_3d_fits_filecontents():
    """Generate the binary contents of a toy FITS file."""
    bytesio = io.BytesIO()
    table = Table(
        [np.arange(12)] * 4, names=['PROB', 'DISTMU', 'DISTSIGMA', 'DISTNORM'])
    table.meta['comment'] = 'This is a comment.'
    table.meta['history'] = 'This is a history line. <This should be escaped.>'
    table.meta['ORDERING'] = 'NESTED'
    with gzip.GzipFile(fileobj=bytesio, mode='wb') as f:
        table.write(f, format='fits')
    return bytesio.getvalue()


def mock_download(filename, graceid):
    if filename == 'test.fits,0' and graceid == 'T12345':
        return toy_3d_fits_filecontents()
    else:
        raise RuntimeError('Asked for unexpected FITS file')


@patch('gwcelery.tasks.gracedb.download.run', mock_download)
@patch('gwcelery.tasks.gracedb.client', autospec=rest.GraceDb)
@patch('ligo.skymap.tool.ligo_skymap_plot.main')
@patch('ligo.skymap.tool.ligo_skymap_plot_volume.main')
def test_annotate_fits(mock_plot_volume, mock_plot, mock_gracedb,
                       toy_3d_fits_filecontents):
    skymaps.annotate_fits('test.fits,0', 'test', 'T12345', ['tag1']).delay(
        toy_3d_fits_filecontents).get()


def test_fits_header(toy_fits_filecontents):
    # Run function under test
    html = skymaps.fits_header(toy_fits_filecontents, 'test.fits')

    # Check output
    assert html == resource_unicode(__name__, 'data/fits_header_result.html')


@patch('ligo.skymap.tool.ligo_skymap_plot.main')
def test_plot_allsky(mock_plot):
    # Run function under test
    skymaps.plot_allsky('')

    # Check that the script would have been run once
    # with the correct arguments
    mock_plot.assert_called_once()
    cmdline, = mock_plot.call_args[0]
    assert cmdline[2].endswith('.png')


def test_is_3d_fits_file(toy_fits_filecontents, toy_3d_fits_filecontents):
    # This is not a 3D FITS file.
    with pytest.raises(ValueError):
        skymaps.is_3d_fits_file(toy_fits_filecontents)
    # This is a 3D FITS file.
    skymaps.is_3d_fits_file(toy_3d_fits_filecontents)


@patch('ligo.skymap.tool.ligo_skymap_plot_volume.main')
def test_plot_volume(mock_plot_volume):
    # Run function under test
    skymaps.plot_volume('')

    # Check that the script would have been run once
    # with the correct arguments
    mock_plot_volume.assert_called_once()
    cmdline, = mock_plot_volume.call_args[0]
    assert cmdline[-2].endswith('.png')



"""Subsystem for sending, receiving, and processing Gamma-ray Coordinates
Network [GCN]_ notices.

References
----------

.. [GCN] https://gcn.gsfc.nasa.gov
"""
import socket
import struct

from celery import Task
from celery.utils.log import get_task_logger
from celery_eternal import EternalProcessTask
from gcn import get_notice_type, NoticeType
import gcn

from ...celery import app
from ..core import DispatchHandler

log = get_task_logger(__name__)

_size_struct = struct.Struct("!I")


class _SendTask(Task):

    def __init__(self):
        self.conn = None


@app.task(queue='voevent', base=_SendTask, ignore_result=True, bind=True,
          autoretry_for=(socket.error,), default_retry_delay=0.001,
          retry_backoff=True, retry_kwargs=dict(max_retries=None),
          shared=False)
def send(self, payload):
    """Send a VOEvent to GCN."""
    payload = payload.encode('utf-8')
    nbytes = len(payload)

    conn = self.conn
    self.conn = None

    if conn is None:
        log.info('creating new socket')
        sock = socket.socket(socket.AF_INET)
        try:
            sock.bind((app.conf['gcn_bind_address'],
                       app.conf['gcn_bind_port']))
            sock.listen(0)
            while True:
                conn, (addr, _) = sock.accept()
                if addr == app.conf['gcn_remote_address']:
                    break
                else:
                    log.error('connection denied to remote host %s', addr)
        finally:
            sock.close()
        conn.setsockopt(socket.SOL_SOCKET, socket.SO_LINGER,
                        struct.pack('ii', 1, 0))
        conn.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, 1)

    log.info('sending payload of %d bytes', nbytes)
    try:
        conn.sendall(_size_struct.pack(nbytes) + payload)
    except:  # noqa
        try:
            conn.shutdown(socket.SHUT_RDWR)
        except:  # noqa
            log.exception('failed to shut down socket')
        conn.close()
        raise
    self.conn = conn


class _VOEventDispatchHandler(DispatchHandler):

    def process_args(self, payload, root):
        notice_type = get_notice_type(root)

        # Just cast to enum for prettier log messages
        try:
            notice_type = NoticeType(notice_type)
        except ValueError:
            pass

        return notice_type, (payload,), {}


handler = _VOEventDispatchHandler()
"""Function decorator to register a handler callback for specified GCN notice
types. The decorated function is turned into a Celery task, which will be
automatically called whenever a matching GCN notice is received.

Parameters
----------
\*keys
    List of GCN notice types to accept
\*\*kwargs
    Additional keyword arguments for :meth:`celery.Celery.task`.

Examples
--------
Declare a new handler like this::

    @gcn.handler(gcn.NoticeType.FERMI_GBM_GND_POS,
                 gcn.NoticeType.FERMI_GBM_FIN_POS)
    def handle_fermi(payload):
        root = lxml.etree.fromstring(payload)
        # do work here...
"""


@app.task(base=EternalProcessTask, shared=False)
def listen():
    """Listen to GCN notices forever. GCN notices are dispatched asynchronously
    to tasks that have been registered with
    :meth:`gwcelery.tasks.gcn.handler`."""
    gcn.listen(handler=handler.dispatch)



from lxml import etree
from urllib.parse import urlparse

from celery.utils.log import get_task_logger

from .. import gcn
from .. import gracedb
from ...celery import app

log = get_task_logger(__name__)


@gcn.handler(gcn.NoticeType.FERMI_GBM_ALERT,
             gcn.NoticeType.FERMI_GBM_FLT_POS,
             gcn.NoticeType.FERMI_GBM_GND_POS,
             gcn.NoticeType.FERMI_GBM_FIN_POS,
             gcn.NoticeType.SWIFT_BAT_GRB_ALERT,
             gcn.NoticeType.SWIFT_BAT_ALARM_LONG,
             gcn.NoticeType.SWIFT_BAT_ALARM_SHORT,
             gcn.NoticeType.SWIFT_BAT_KNOWN_SRC,
             gcn.NoticeType.SWIFT_BAT_GRB_LC,
             gcn.NoticeType.SNEWS,
             queue='exttrig',
             shared=False)
def handle(payload):
    """Handles the payload from the Fermi, Swift and SNEWS alerts.
    Prepares the alert to be sent to graceDB as 'E' events."""
    event_type = app.conf.external_trigger_event_type
    root = etree.fromstring(payload)
    u = urlparse(root.attrib['ivorn'])
    stream_path = u.path

    #  Get TrigID
    trig_id = root.find("./What/Param[@name='TrigID']").attrib['value']

    stream_obsv_dict = {'/SWIFT': 'Swift',
                        '/Fermi': 'Fermi',
                        '/SNEWS': 'SNEWS'}
    event_observatory = stream_obsv_dict[stream_path]
    query = 'group: {} pipeline: {} grbevent.trigger_id = "{}"'.format(
        event_type, event_observatory, trig_id)
    events = gracedb.get_events(query=query)

    if events:
        assert len(events) == 1, 'Found more than one matching GraceDb entry'
        event, = events
        graceid = event['graceid']
        gracedb.replace_event(graceid, payload)

    else:
        gracedb.create_event(filecontents=payload,
                             search='GRB',
                             group=event_type,
                             pipeline=event_observatory)



"""Routing of LVAlert messages to other tasks."""
from urllib.error import URLError

from celery import group

from ..celery import app
# from . import circulars
from . import gracedb
from . import lvalert
from . import raven
from . import skymaps


@lvalert.handler('superevent',
                 'test_superevent',
                 shared=False)
def handle(alert):
    """Schedule annotations for new superevents.

    This task calls
    :func:`~gwcelery.tasks.orchestrator.annotate_superevent` after a timeout
    specified by the :obj:`~gwcelery.celery.Base.orchestrator_timeout`
    configuration variable.
    """

    if alert['alert_type'] != 'new':
        return

    superevent_id = alert['object']['superevent_id']
    (
        get_preferred_event.s(superevent_id) |
        annotate_superevent.s(superevent_id)
    ).apply_async(countdown=app.conf['orchestrator_timeout'])


@lvalert.handler('superevent',
                 'external_fermi',
                 'external_fermi_grb',
                 'external_grb',
                 'external_snews',
                 'external_snews_supernova',
                 'external_swift',
                 shared=False)
def handle_superevents_externaltriggers(alert):
    """Parse an LVAlert message related to superevents/external triggers and
    dispatch it to other tasks."""
    # Determine GraceDb ID
    graceid = alert['uid']

    if alert['alert_type'] == 'new':
        raven.coincidence_search(graceid, alert['object']).delay()


@app.task(autoretry_for=(URLError, TimeoutError), default_retry_delay=20.0,
          retry_backoff=True, retry_kwargs=dict(max_retries=500), shared=False)
def download(*args, **kwargs):
    """Download a file from GraceDb.

    This works just like :func:`gwcelery.tasks.gracedb.download`, except that
    it is retried for both `TimeoutError` and `URLError`. In particular, it
    will be retried for 404 (not found) errors.

    FIXME: Is there a way to change retry settings when calling a task,
    instead of creating an entirely new task?"""
    return gracedb.download(*args, **kwargs)


@app.task(autoretry_for=(URLError, TimeoutError), default_retry_delay=1.0,
          retry_backoff=True, retry_backoff_max=10.0,
          retry_kwargs=dict(max_retries=10), shared=False)
def get_preferred_event(superevent_id):
    """Determine preferred event for a superevent by querying GraceDb.

    This works just like :func:`gwcelery.tasks.gracedb.get_superevent`, except
    that it returns only the preferred event, and not the entire GraceDb JSON
    response."""
    return gracedb.get_superevent(superevent_id)['preferred_event']


@app.task(ignore_result=True, shared=False)
def annotate_superevent(preferred_event_id, superevent_id):
    """Perform annotations for a new superevent."""
    (
        download.s('bayestar.fits.gz', preferred_event_id)
        |
        group(
            gracedb.upload.s(
                'bayestar.fits.gz',
                superevent_id,
                message='BAYESTAR localization copied from {}'.format(
                    preferred_event_id),
                tags=['sky_loc', 'lvem']
            ),

            skymaps.annotate_fits(
                'bayestar.fits.gz',
                'bayestar',
                superevent_id,
                ['sky_loc', 'lvem']
            )
        )
        |
        gracedb.create_voevent.si(
            superevent_id, 'preliminary',
            skymap_type='BAYESTAR',
            skymap_filename='bayestar.fits.gz',
            skymap_image_filename='bayestar.png'
        )
        # FIXME: Circulars don't work yet due to a regression in ligo-gracedb.
        # See https://git.ligo.org/lscsoft/gracedb-client/issues/7
        #
        # |
        # circulars.create_circular.si(superevent_id)
        # |
        # gracedb.upload.s(
        #     'circular.txt',
        #     superevent_id,
        #     'Automated circular'
        # )
    ).delay()



"""Validate LIGO/Virgo GCN notices to make sure that their contents match the
original VOEvent notices that we sent."""
from urllib.parse import urlparse

import lxml.etree

from .. import gcn
from .. import gracedb


@gcn.handler(gcn.NoticeType.LVC_PRELIMINARY,
             gcn.NoticeType.LVC_INITIAL,
             gcn.NoticeType.LVC_UPDATE,
             shared=False)
def validate_voevent(payload):
    """Check that the contents of a public LIGO/Virgo GCN matches the original
    VOEvent in GraceDB."""
    root = lxml.etree.fromstring(payload)

    # Which GraceDB ID does this refer to?
    graceid = root.find("./What/Param[@name='GraceID']").attrib['value']

    # Which VOEvent does this refer to?
    u = urlparse(root.attrib['ivorn'])
    assert u.scheme == 'ivo', (
        'IVORN has unexpected scheme: {!r}'.format(u.scheme))
    assert u.netloc == 'nasa.gsfc.gcn', (
        'IVORN has unexpected netloc: {!r}'.format(u.netloc))
    assert u.path == '/LVC', (
        'IVORN has unexpected path: {!r}'.format(u.path))
    local_id = u.fragment
    filename = local_id + '.xml'

    # Download and parse original VOEvent
    orig = lxml.etree.fromstring(gracedb.download(filename, graceid))

    xpath = ".//Param[@name='{}']"
    for orig_name, root_name in [
            ['skymap_fits_shib', 'SKYMAP_URL_FITS_SHIB'],
            ['skymap_fits_x509', 'SKYMAP_URL_FITS_X509'],
            ['skymap_fits_basic', 'SKYMAP_URL_FITS_BASIC'],
            ['skymap_png_shib', 'SKYMAP_URL_PNG_SHIB'],
            ['skymap_png_x509', 'SKYMAP_URL_PNG_X509'],
            ['skymap_png_basic', 'SKYMAP_URL_PNG_BASIC']]:

        orig_elem = orig.find(xpath.format(orig_name))
        root_elem = root.find(xpath.format(root_name))

        if orig_elem is None:
            assert root_elem is None, (
                'GCN has unexpected parameter: {!r}'.format(root_name))
        else:
            assert root_elem is not None, (
                'GCN is missing parameter: {!r}'.format(root_name))
            orig_value = orig_elem.attrib.get('value')
            root_value = root_elem.attrib.get('value')
            assert root_value == orig_value, (
                'GCN parameter {!r} has value {!r}, but '
                'original VOEvent parameter {!r} '
                'has value {!r}'.format(
                    root_name, root_value, orig_name, orig_value))

    # Find matching GraceDB log entry
    log = gracedb.get_log(graceid)
    entry, = (e for e in log if e['filename'] == filename)
    log_number = entry['N']

    # Tag the VOEvent to indicate that it was received correctly
    gracedb.create_tag('gcn_received', log_number, graceid)



from unittest.mock import patch
from xml.sax import SAXParseException

from astropy import table
from ligo.gracedb import rest
import numpy as np
import pkg_resources
import pytest

from ..tasks.bayestar import handle, localize
from . import resource_json


def mock_download(filename, graceid, *args, **kwargs):
    assert graceid == 'T250822'
    if filename == 'coinc.xml':
        return pkg_resources.resource_string(__name__, 'data/coinc.xml')
    elif filename == 'psd.xml.gz':
        return pkg_resources.resource_string(__name__, 'data/psd.xml.gz')
    else:
        raise ValueError


@patch('gwcelery.tasks.gracedb.download.run', mock_download)
@patch('gwcelery.tasks.bayestar.localize.run')
@patch('gwcelery.tasks.gracedb.client', autospec=rest.GraceDb)
def test_handle(mock_gracedb, mock_localize):
    """Test that an LVAlert message for a newly uploaded PSD file triggers
    BAYESTAR."""
    alert = resource_json(__name__, 'data/lvalert_psd.json')
    handle(alert)
    mock_localize.assert_called_once()


@patch('gwcelery.tasks.gracedb.download.run', mock_download)
@patch('gwcelery.tasks.bayestar.localize.run')
@patch('gwcelery.tasks.gracedb.client', autospec=rest.GraceDb)
def test_handle_ignored(mock_gracedb, mock_localize):
    """Test that unrelated LVAlert messages do not trigger BAYESTAR."""
    alert = resource_json(__name__, 'data/lvalert_detchar.json')
    handle(alert)
    mock_localize.assert_not_called()


@patch('gwcelery.tasks.gracedb.client', autospec=rest.GraceDb)
def test_localize_bad_psd(mock_gracedb):
    """Test running BAYESTAR with a pad PSD file"""
    # Test data
    coinc = pkg_resources.resource_string(__name__, 'data/coinc.xml')
    psd = b''

    # Run function under test
    with pytest.raises(SAXParseException):
        localize((coinc, psd), 'G211117')


def mock_bayestar(*args, **kwargs):
    return table.Table({'UNIQ': np.arange(4, 16, dtype=np.uint64),
                        'PROBDENSITY': np.ones(12),
                        'DISTMU': np.ones(12),
                        'DISTSIGMA': np.ones(12),
                        'DISTNORM': np.ones(12)})


@pytest.mark.parametrize('disabled_detectors', [None,
                                                ['H1', 'L1'],
                                                ['H1', 'L1', 'V1']])
@patch('gwcelery.tasks.gracedb.client', autospec=rest.GraceDb)
@patch('ligo.skymap.bayestar.localize', mock_bayestar)
def test_localize(mock_gracedb, disabled_detectors):
    """Test running BAYESTAR on G211117"""
    # Test data
    coinc = pkg_resources.resource_string(__name__, 'data/coinc.xml')
    psd = pkg_resources.resource_string(__name__, 'data/psd.xml.gz')

    # Run function under test
    fitscontent = localize(
        (coinc, psd), 'G211117', disabled_detectors=disabled_detectors)

    # FIXME: should do some sanity checks of the sky map here
    assert fitscontent



"""LVAlert client."""
import json
import netrc
import os
import uuid

from celery_eternal import EternalTask
from celery.utils.log import get_task_logger
from pyxmpp2.client import Client
from pyxmpp2.interfaces import (EventHandler, event_handler,
                                message_stanza_handler, QUIT,
                                XMPPFeatureHandler)
from pyxmpp2.jid import JID
from pyxmpp2.mainloop.interfaces import TimeoutHandler, timeout_handler
from pyxmpp2.settings import XMPPSettings
from pyxmpp2.streamevents import DisconnectedEvent

from ..celery import app
from .core import DispatchHandler
from . import gracedb

log = get_task_logger(__name__)

ns = {'ns1': 'http://jabber.org/protocol/pubsub#event',
      'ns2': 'http://jabber.org/protocol/pubsub'}


class _LVAlertDispatchHandler(DispatchHandler):

    def process_args(self, node, payload):
        # Determine GraceDB service URL
        alert = json.loads(payload)
        try:
            try:
                self_link = alert['object']['links']['self']
            except KeyError:
                self_link = alert['object']['self']
        except KeyError:
            log.exception('LVAlert message does not contain an API URL: %r',
                          alert)
            return None, None, None
        base, api, _ = self_link.partition('/api/')
        service = base + api

        if service != gracedb.client.service_url:
            log.warn('ignoring LVAlert message because it is intended for '
                     'GraceDb server %s, but we are set up for server %s',
                     service, gracedb.client.service_url)
            return None, None, None

        return super().process_args(node, alert)


handler = _LVAlertDispatchHandler()
"""Function decorator to register a handler callback for specified LVAlert
message types. The decorated function is turned into a Celery task, which will
be automatically called whenever a matching LVAlert message is received.

Parameters
----------
\*keys
    List of LVAlert message types to accept
\*\*kwargs
    Additional keyword arguments for :meth:`celery.Celery.task`.

Examples
--------
Declare a new handler like this::

    @lvalert.handler('cbc_gstlal',
                     'cbc_pycbc',
                     'cbc_mbta')
    def handle_cbc(alert_content):
        # do work here...
"""


def _handle_messages(xml):
    for node in xml.iterfind('.//ns1:items[@node]', ns):
        for entry in node.iterfind('.//ns2:entry', ns):
            handler.dispatch(node.attrib['node'], entry.text)


class _LVAlertClient(EventHandler, TimeoutHandler, XMPPFeatureHandler):

    def __init__(self, server, task=None):
        # Look up username and password from the netrc file.
        netrcfile = os.environ.get('NETRC')
        auth = netrc.netrc(netrcfile).authenticators(server)
        if auth is None:
            raise RuntimeError('No matching netrc entry found')
        username, _, password = auth

        # Create a JID with a unique resource name.
        resource = uuid.uuid4().hex
        jid = JID(username + '@' + server + '/' + resource)

        settings = XMPPSettings(dict(
            starttls=True, tls_verify_peer=False, password=password))

        self.__task = task
        self.__client = Client(jid, [self], settings)
        self.__client.main_loop.add_handler(self)

    @event_handler(DisconnectedEvent)
    def __handle_disconnected(self, *args, **kwargs):
        return QUIT

    @message_stanza_handler()
    def __handle_message(self, stanza):
        log.info('Got message')
        _handle_messages(stanza.as_xml())
        return True

    @timeout_handler(1, recurring=True)
    def __handle_timeout(self, *args, **kargs):
        if self.__task is not None and self.__task.is_aborted():
            log.info('Task aborted, quitting event loop')
            self.__client.disconnect()
        return True

    def run(self):
        log.info('Connecting to %r', str(self.__client.jid))
        self.__client.connect()
        self.__client.run()
        log.info('Reached end of main loop')


@app.task(base=EternalTask, bind=True, shared=False)
def listen(self):
    """Listen for LVAlert messages forever. LVAlert messages are dispatched
    asynchronously to tasks that have been registered with
    :meth:`gwcelery.tasks.lvalert.handler`."""
    _LVAlertClient(app.conf['lvalert_host'], self).run()



import json
import logging
import socket
from threading import Thread
from time import sleep
from unittest.mock import MagicMock, patch

import gcn
from gcn.voeventclient import _recv_packet
import lxml.etree
import pkg_resources
import pytest

from ..tasks.gcn import handler, listen, send
from .. import app

# Test data
with pkg_resources.resource_stream(
        __name__, 'data/lvalert_voevent.json') as f:
    lvalert = json.load(f)
voevent = lvalert['object']['text']


@pytest.fixture
def send_thread():
    thread = Thread(target=send, args=(voevent,))
    thread.daemon = True
    thread.start()
    sleep(1)
    yield thread
    thread.join()
    if send.conn is not None:
        try:
            send.conn.close()
        except socket.error:
            pass
        send.conn = None


@pytest.mark.enable_socket
def test_send_connection_closed(send_thread):
    """Test sending a VOEvent over loopback to a connection that
    is immediately closed."""
    sock = socket.socket(socket.AF_INET)
    sock.connect(('127.0.0.1', 53410))
    sock.shutdown(socket.SHUT_RDWR)
    sock.close()


@pytest.mark.enable_socket
def test_send(send_thread):
    """Test sending a VOEvent over loopback."""
    # First, simulate connecting from a disallowed IP address.
    # The connection should be refused.
    app.conf['gcn_remote_address'] = '192.0.2.0'
    sock = socket.socket(socket.AF_INET)
    try:
        sock.settimeout(0.1)
        with pytest.raises(socket.error):
            sock.connect(('127.0.0.1', 53410))
            packet = _recv_packet(sock)
    finally:
        try:
            sock.shutdown(socket.SHUT_RDWR)
        except socket.error:
            pass
        sock.close()

    # Now, simulate connecting from the allowed IP address.
    # The VOEvent should be received.
    app.conf['gcn_remote_address'] = '127.0.0.1'
    sock = socket.socket(socket.AF_INET)
    try:
        sock.settimeout(0.1)
        sock.connect(('127.0.0.1', 53410))
        packet = _recv_packet(sock)
    finally:
        try:
            sock.shutdown(socket.SHUT_RDWR)
        except socket.error:
            pass
        sock.close()
    assert packet == voevent.encode('utf-8')


@pytest.mark.enable_socket
def test_listen(monkeypatch):
    """Test that the listen task would correctly launch gcn.listen()."""
    mock_gcn_listen = MagicMock()
    monkeypatch.setattr('gcn.listen', mock_gcn_listen)
    listen.run()
    mock_gcn_listen.assert_called_once()


def fake_gcn(notice_type):
    # Check the real GCN notice, which is valid.
    payload = pkg_resources.resource_string(
        __name__, 'data/G298048-1-Initial.gcn.xml')
    root = lxml.etree.fromstring(payload)
    notice_type = str(int(notice_type))
    root.find(".//Param[@name='Packet_Type']").attrib['value'] = notice_type
    return lxml.etree.tostring(root), root


def test_unrecognized_notice_type(caplog):
    """Test handling an unrecognized (enum not defined) notice type."""
    caplog.set_level(logging.WARNING)
    handler.dispatch(*fake_gcn(10000))
    record, = caplog.records
    assert record.message == 'ignoring unrecognized key: 10000'


def test_unregistered_notice_type(caplog):
    """Test handling an unregistered notice type."""
    caplog.set_level(logging.WARNING)
    handler.dispatch(*fake_gcn(gcn.NoticeType.SWIFT_UVOT_POS_NACK))
    record, = caplog.records
    assert record.message == ('ignoring unrecognized key: '
                              '<NoticeType.SWIFT_UVOT_POS_NACK: 89>')


@pytest.fixture
def reset_handlers():
    old_handler = dict(handler)
    handler.clear()
    yield
    handler.update(old_handler)


def test_registered_notice_type(reset_handlers):
    @handler(gcn.NoticeType.AGILE_POINTDIR, gcn.NoticeType.AGILE_TRANS)
    def agile_handler(payload):
        pass

    with patch.object(agile_handler, 'run') as mock_run:
        handler.dispatch(*fake_gcn(gcn.NoticeType.SWIFT_UVOT_POS_NACK))
        mock_run.assert_not_called()
        handler.dispatch(*fake_gcn(gcn.NoticeType.AGILE_POINTDIR))
        mock_run.assert_called_once()



"""Annotations for sky maps."""
from astropy.io import fits
from celery import group
from ligo.skymap.tool import ligo_skymap_plot
from ligo.skymap.tool import ligo_skymap_plot_volume

from . import gracedb
from ..celery import app
from ..jinja import env
from ..util.tempfile import NamedTemporaryFile


def annotate_fits(versioned_filename, filebase, graceid, tags):
    """Perform annotations on a sky map.

    This function downloads a FITS file and then generates and uploads all
    derived images as well as an HTML dump of the FITS header.
    """
    header_msg = (
        'FITS headers for <a href="/apiweb/events/{graceid}/files/'
        '{versioned_filename}">{versioned_filename}</a>').format(
            graceid=graceid, versioned_filename=versioned_filename)
    allsky_msg = (
        'Mollweide projection of <a href="/apiweb/events/{graceid}/files/'
        '{versioned_filename}">{versioned_filename}</a>').format(
            graceid=graceid, versioned_filename=versioned_filename)
    volume_msg = (
        'Volume rendering of <a href="/apiweb/events/{graceid}/files/'
        '{versioned_filename}">{versioned_filename}</a>').format(
            graceid=graceid, versioned_filename=versioned_filename)

    return group(
        fits_header.s(versioned_filename) |
        gracedb.upload.s(
            filebase + '.html', graceid, header_msg, tags),

        plot_allsky.s() |
        gracedb.upload.s(
            filebase + '.png', graceid, allsky_msg, tags),

        is_3d_fits_file.s() |
        plot_volume.s() |
        gracedb.upload.s(
            filebase + '.volume.png', graceid, volume_msg, tags)
    )


@app.task(shared=False)
def fits_header(filecontents, filename):
    """Dump FITS header to HTML."""
    template = env.get_template('fits_header.html')
    with NamedTemporaryFile(content=filecontents) as fitsfile, \
            fits.open(fitsfile.name) as hdus:
        return template.render(filename=filename, hdus=hdus)


@app.task(shared=False)
def plot_allsky(filecontents):
    """Plot a Mollweide projection of a sky map."""
    with NamedTemporaryFile(mode='rb', suffix='.png') as pngfile, \
            NamedTemporaryFile(content=filecontents) as fitsfile:
        ligo_skymap_plot.main([fitsfile.name, '-o', pngfile.name,
                               '--annotate', '--contour', '50', '90'])
        return pngfile.read()


@app.task(shared=False, throws=(ValueError,))
def is_3d_fits_file(filecontents):
    """Determine if a FITS file has distance information. If it does, then
    the file contents are returned. If it does not, then a :obj:`ValueError` is
    raised."""
    try:
        with NamedTemporaryFile(content=filecontents) as fitsfile:
            if fits.getval(fitsfile.name, 'TTYPE4', 1) == 'DISTNORM':
                return filecontents
    except (KeyError, IndexError):
        raise ValueError('Not a 3D FITS file')


@app.task(queue='openmp', shared=False)
def plot_volume(filecontents):
    """Plot a Mollweide projection of a sky map."""
    with NamedTemporaryFile(mode='rb', suffix='.png') as pngfile, \
            NamedTemporaryFile(content=filecontents) as fitsfile:
        ligo_skymap_plot_volume.main([fitsfile.name, '-o',
                                      pngfile.name, '--annotate'])
        return pngfile.read()



"""Celery application configuration."""

from celery import Celery


class Base:
    """Base application configuration."""

    # Celery application settings.
    # Use pickle serializer, because it supports byte values.

    accept_content = ['json', 'pickle']
    event_serializer = 'json'
    result_serializer = 'pickle'
    task_serializer = 'pickle'

    # GWCelery-specific settings.

    gcn_bind_address = ''
    """Address on which to listen for outbound (sending) connections to GCN;
    empty to listen on all addresses."""

    gcn_bind_port = 5341
    """Port on which to listen for outbound (sending) connections to GCN."""

    gcn_remote_address = 'capella2.gsfc.nasa.gov'
    """Address to listen for inbound (receiving) connections to GCN."""

    superevent_d_t_start = dict(gstlal=1.0,
                                pycbc=1.0,
                                mbtaonline=1.0)
    """Pipeline based lower extent of superevent segments.
    For cwb and lib this is decided from extra attributes."""

    superevent_d_t_end = dict(gstlal=1.0,
                              pycbc=1.0,
                              mbtaonline=1.0)
    """Pipeline based upper extent of superevent segments
    For cwb and lib this is decided from extra attributes."""

    superevent_query_d_t_start = 100.
    """Lower extent of superevents query"""

    superevent_query_d_t_end = 100.
    """Upper extent of superevents query"""

    superevent_default_d_t_start = 1.0
    """Default lower extent of superevent segments"""

    superevent_default_d_t_end = 1.0
    """Default upper extent for superevent segments"""

    superevent_far_threshold = 1.9e-07
    """Maximum false alarm rate to consider events superevents."""

    orchestrator_timeout = 60.0
    """The orchestrator will wait this many seconds from the time of the
    creation of a new superevent to the time that annotations begin, in order
    to let the superevent manager's decision on the preferred event
    stabilize."""


class Production(Base):
    """Application configuration for ``gracedb.ligo.org``."""

    external_trigger_event_type = 'External'
    """GraceDb group for external triggers (e.g. GRBs, SNe, neutrinos)."""

    lvalert_host = 'lvalert.cgca.uwm.edu'
    """LVAlert host."""

    gracedb_host = 'gracedb.ligo.org'
    """GraceDb host."""


class Test(Base):
    """Application configuration for ``gracedb-test.ligo.org``."""

    external_trigger_event_type = 'Test'
    """GraceDb group for external triggers (e.g. GRBs, SNe, neutrinos)."""

    lvalert_host = 'lvalert-test.cgca.uwm.edu'
    """LVAlert host."""

    gracedb_host = 'gracedb-test.ligo.org'
    """GraceDb host."""


class Development(Test):
    """Application configuration for ``gracedb-dev1.ligo.org``."""

    gracedb_host = 'gracedb-dev1.ligo.org'
    """GraceDb host."""


class Playground(Test):
    """Application configuration for ``gracedb-playground.ligo.org``."""

    gracedb_host = 'gracedb-playground.ligo.org'
    """GraceDb host."""


# Celery application object.
# Use redis broker, because it supports locks (and thus singleton tasks).
app = Celery('gwcelery', broker='redis://', config_source=Playground)

# Use the same URL for both the result backend and the broker.
app.conf['result_backend'] = app.conf.broker_url



"""Communication with GraceDB."""
import os
import sys

from ligo.gracedb import rest
from celery.local import PromiseProxy
from celery.utils.log import get_task_logger

from ..celery import app

# Defer initializing the GraceDb REST client until it is needed,
# because if the user lacks valid credentials, then the API will
# raise an exception as soon as it is instantiated---which would
# otherwise make it impossible to import this module without first
# logging in.
#
# FIXME: sphinx gets confused by PromiseProxy and prints:
#     Recursion error:
#     maximum recursion depth exceeded
# so don't create the proxy if we are being run by sphinx.
prog = os.path.basename(sys.argv[0])
if prog != 'sphinx-build' and 'build_sphinx' not in sys.argv:
    client = PromiseProxy(rest.GraceDb,
                          ('https://' + app.conf.gracedb_host + '/api/',))

log = get_task_logger(__name__)


def task(*args, **kwargs):
    return app.task(*args, **kwargs, autoretry_for=(TimeoutError,),
                    default_retry_delay=20.0, retry_backoff=True,
                    retry_kwargs=dict(max_retries=10))


@task(shared=False)
def create_event(filecontents, search, pipeline, group):
    """Create an event in GraceDb."""
    response = client.createEvent(group=group, pipeline=pipeline,
                                  filename='initial.data', search=search,
                                  filecontents=filecontents)
    return response.json()['graceid']


@task(ignore_result=True, shared=False)
def create_tag(tag, n, graceid):
    """Create a tag in GraceDb."""
    client.createTag(graceid, n, tag)


@task(shared=False)
def create_voevent(graceid, voevent_type, **kwargs):
    response = client.createVOEvent(graceid, voevent_type, **kwargs)
    response = response.json()
    # FIXME: createVOEvent() no longer returns the VOEvent content.
    # See https://git.ligo.org/lscsoft/gracedb/issues/13
    # response = response['text']
    return response


@task(shared=False)
def download(filename, graceid):
    """Download a file from GraceDB."""
    return client.files(graceid, filename, raw=True).read()


@task(shared=False)
def get_events(query=None, orderby=None, count=None, columns=None):
    """Get events from GraceDb."""
    return list(client.events(query=query, orderby=orderby,
                count=count, columns=columns))


@task(shared=False)
def get_event(graceid):
    """Retrieve an event from GraceDb."""
    return client.event(graceid).json()


@task(shared=False)
def get_log(graceid):
    """Get all log messages for an event in GraceDb."""
    return client.logs(graceid).json()['log']


@task(shared=False)
def get_superevent(graceid):
    """Retrieve a superevent from GraceDb."""
    return client.superevent(graceid).json()


@task(shared=False)
def replace_event(graceid, payload):
    """Get an event from GraceDb."""
    client.replaceEvent(graceid, 'initial.data', filecontents=payload)


@task(ignore_result=True, shared=False)
def upload(filecontents, filename, graceid, message, tags=None):
    """Upload a file to GraceDB."""
    # FIXME: it would be more elegant to have `tags=()` in the kwargs, but
    # gracedb-client does not understand tuples of tags.
    # See https://git.ligo.org/lscsoft/gracedb-client/issues/5
    if tags is None:
        tags = []
    client.writeLog(graceid, message, filename, filecontents, tags)


@app.task(shared=False)
def get_superevents(gid, query=None):
    """Iterate through superevents in gracedb and return sid if
    gid exists in the association.

    Parameters
    ----------
    gid : str
        uid of the trigger to be checked

    query : str
        optional query to be passed to :meth:`superevents`

    Returns
    -------
    superevent_id : str
        uid of the superevent. None if not found
    preferred_flag : bool
        True if gid is found and it is preferred. None if not found.
    superevents : list
        The list of the superevents.
    """
    superevents = list(client.superevents(query=query, orderby='t_0'))
    for superevent in superevents:
        preferred_flag = False
        # check preferred_event first
        if gid == superevent['preferred_event']:
            preferred_flag = True
            log.info('Found association (Preferred) %s <-> %s',
                     gid, superevent['superevent_id'])
            return superevent['superevent_id'], preferred_flag, superevents
        # then check the gw_events
        elif gid in superevent['gw_events']:
            log.info('Found association (NOT Preferred) %s <-> %s',
                     gid, superevent['superevent_id'])
            return superevent['superevent_id'], preferred_flag, superevents
    return None, False, superevents


@task(ignore_result=True, shared=False)
def update_superevent(superevent_id, t_start=None,
                      t_end=None, t_0=None, preferred_event=None):
    """
    Update superevent information. Wrapper around
    :meth:`updateSuperevent`

    Parameters
    ----------
    superevent_id : str
        superevent uid
    t_start : float
        start of superevent time window, unchanged if None
    t_end : float
        end of superevent time window, unchanged if None
    t_0 : float
        superevent t_0, unchanged if None
    preferred_event : str
        uid of the preferred event, unchanged if None
    """
    client.updateSuperevent(superevent_id, t_start=t_start, t_end=t_end,
                            t_0=t_0, preferred_event=preferred_event)


@task(ignore_result=True, shared=False)
def create_superevent(graceid, t0, d_t_start, d_t_end):
    """Create new superevent in GraceDb with `graceid`

    Parameters
    ----------
    graceid : str
        graceid with which superevent is created.
    t0 : float
        `t_0` parameter of superevent
    d_t_start : float
        superevent `t_start` = `t0 - d_t_start`
    d_t_end : float
        superevent `t_end` = `t0 + t_end`
    """
    ts = t0 - d_t_start
    te = t0 + d_t_end
    client.createSuperevent(ts, t0, te, preferred_event=graceid)


@task(ignore_result=True, shared=False)
def add_event_to_superevent(superevent_id, graceid):
    """Add an event to a superevent in GraceDb."""
    client.addEventToSuperevent(superevent_id, graceid)



"""Rapid sky localization with BAYESTAR."""
import io
import logging
import os
import tempfile

from celery import group
from celery.exceptions import Ignore
from ligo.gracedb.logging import GraceDbLogHandler
from ligo.skymap import bayestar as _bayestar
from ligo.skymap.io import events
from ligo.skymap.io import fits

from ..celery import app
from . import lvalert
from . import gracedb

log = logging.getLogger('BAYESTAR')


@lvalert.handler('cbc_gstlal',
                 'cbc_pycbc',
                 'cbc_mbtaonline',
                 'test_gstlal',
                 'test_pycbc',
                 'test_mbtaonline',
                 shared=False)
def handle(alert):
    """Peform end-to-end rapid sky localization with BAYESTAR, including
    GraceDB downloads and uploads.

    This function downloads all of the required inputs for BAYESTAR, runs rapid
    sky localization, and uploads the resulting sky map FITS files to GraceDB.

    Internally, all of the heavy lifting is done by :meth:`localize`.
    """

    # Skip alerts that are not meant for us.
    if alert['alert_type'] != 'update' or alert.get('file') != 'psd.xml.gz':
        return

    graceid = alert['uid']

    (
        group(
            gracedb.download.s('coinc.xml', graceid),
            gracedb.download.s('psd.xml.gz', graceid)
        )
        |
        localize.s(graceid)
        |
        gracedb.upload.s(
            'bayestar.fits.gz', graceid,
            'sky localization complete', ['sky_loc', 'lvem']
        )
    ).delay()


@app.task(queue='openmp', shared=False)
def localize(coinc_psd, graceid, filename='bayestar.fits.gz',
             disabled_detectors=None):
    """Do the heavy lifting of generating a rapid localization using BAYESTAR.

    This function runs the computationally-intensive part of BAYESTAR.
    The `coinc.xml` and `psd.xml.gz` files should already have been downloaded
    by :func:`bayestar`.

    This task should execute in a special queue for computationally intensive
    OpenMP parallel tasks.
    """
    handler = GraceDbLogHandler(gracedb.client, graceid)
    handler.setLevel(logging.INFO)
    log.addHandler(handler)

    try:
        # A little bit of Cylon humor
        log.info('by your command...')

        # Parse event
        coinc, psd = coinc_psd
        coinc = io.BytesIO(coinc)
        psd = io.BytesIO(psd)
        event_source = events.ligolw.open(coinc, psd_file=psd, coinc_def=None)
        if disabled_detectors:
            event_source = events.detector_disabled.open(
                event_source, disabled_detectors)
        event, = event_source.values()

        # Run BAYESTAR
        log.info("starting sky localization")
        skymap = _bayestar.rasterize(_bayestar.localize(event))
        skymap.meta['objid'] = str(graceid)
        log.info("sky localization complete")

        with tempfile.TemporaryDirectory() as tmpdir:
            fitspath = os.path.join(tmpdir, filename)
            fits.write_sky_map(fitspath, skymap, nest=True)
            return open(fitspath, 'rb').read()
    except events.DetectorDisabledError:
        raise Ignore()
    except:  # noqa
        # Produce log message for any otherwise uncaught exception
        log.exception("sky localization failed")
        raise
    finally:
        log.removeHandler(handler)
        del handler
